import argparse
import sys
import json
from tqdm import tqdm

from msorm import models


def init(*args):
    if len(args) >= 1:
        parser = argparse.ArgumentParser(description='Argparse Test script')

        parser.add_argument("-ip", help="ip address")
        parser.add_argument("-db", help="database name")
        parser.add_argument("-u", help="username")
        parser.add_argument("-p", help="password")
        parser.add_argument("-f", help="file name", default="models.py")
        args_ = parser.parse_args(list(args))
        models.init(args_.ip, args_.db, args_.u, args_.p)
        file_name = args_.f
        ip, database, username, password = args_.ip, args_.db, args_.u, args_.p

    else:
        print("Welcome to msorm database initializer")
        while 1:
            ip = input("Please enter the database ip address: ")
            database = input("Please enter the database name: ")
            username = input("Please enter the username: ")
            password = input("Please enter the password: ")
            file_name = input("Please enter the output file name[Default: models.py]: ")
            file_name = file_name if file_name != "" else "models.py"
            print(
                f"database ip: {ip}\ndatabase name: {database}\n database username: {username}\n password: {password}")
            answer = input("Are all values are valid?[Y/n]: ")
            if answer.lower() == 'n':
                print("Please re-enter required fields!")
                continue
            else:
                break
        models.init(ip, database, username, password)
    intend = "    "
    models_py = \
        f"""from msorm import models
from msorm.models import Field
models.init("{ip}", "{database}", "{username}", "{password}")
"""
    crsr = models.connection.cursor()
    table_names = [x[2] for x in crsr.tables(tableType='TABLE')]
    table_names.remove("sysdiagrams")
    tables = {k: {} for k in table_names}
    table_scores = {}
    for table in tqdm(table_names):
        table_scores[table] = 0
        primarykey = False
        for row in crsr.columns(table=table):
            if row.type_name == "int identity":
                primarykey = True
                type_name = "primaryKey"
            else:
                type_name = row.type_name

            is_nullable = True if row.is_nullable == "YES" else False
            __primaryKey__ = False

            tables[table][row.column_name] = f"Field.{type_name}(null={is_nullable})"
        if not primarykey:
            tables[table]["__primaryKey__"] = __primaryKey__

    for table in tqdm(table_names):
        for fk in crsr.foreignKeys(table=table):
            table_scores[table]+=1
            pk_column_name = fk.pkcolumn_name
            fk_table_name = fk.fktable_name
            tables[fk_table_name][pk_column_name] = f"Field.foreignKey(model={fk.pktable_name},name='{pk_column_name}')"
    table_scores = [[k,v] for k,v in table_scores.items()]
    table_scores.sort(key=lambda x: x[1],reverse=True)
    table_scores = [t[0] for t in table_scores]
    tables = {k:tables[k] for k in table_scores}

    for last_table, fields in tables.items():
        models_py += f"\n\nclass {last_table}(models.Model):"
        for field, val in fields.items():
            if field == "__primaryKey__" and val == False:
                models_py += f"\n{intend}{field} = False"
                continue
            models_py += f"\n{intend}{field} = {val}"
    with open(file_name, "w", encoding="utf-8") as f:
        f.write(models_py)
    with open("secrets2.json", "w") as fi:
        json.dump(tables,fi,indent=2)


processes = {"init": init}
if __name__ == '__main__':
    args = sys.argv[1:]
    processes.get(args.pop(0))(*args)



import datetime
import sys

import pyodbc


class __MFL__:
    """mssql_field_limits"""
    # region exact numeric fields
    bit_max = 1
    bit_min = 0
    bigint_max = 9_223_372_036_854_775_807
    bigint_min = -bigint_max
    int_max = 2_147_483_647
    int_min = -int_max
    smallint_max = 32_767
    smallint_min = -smallint_max
    tinyint_max = 255
    tinyint_min = 0
    decimal_max = float("inf")
    decimal_min = -float("inf")
    numeric_max = float("inf")
    numeric_min = -float("inf")
    # TODO: Add support for money symbols and convertion
    money_max = 922_337_203_685_477.58
    money_min = -money_max
    smallmoney_max = 214_748.3647
    smallmoney_min = -smallmoney_max
    # endregion
    # region float and real

    float_max = real_max = float("inf")
    float_min = real_min = float("-inf")

    # endregion
    # region String types mostly
    char_max = 8000
    char_min = 1
    nchar_max = 4000
    nchar_min = 1
    varchar_max = 8000
    varchar_min = 1
    nvarchar_max = 4000
    nvarchar_min = 1
    text_max = 2_147_483_647
    text_min = 1
    ntext_max = 2_147_483_647
    ntext_min = 1

    # endregion
    # region binary
    # min and max depends on byte size
    binary_min = varbinary_min = 1
    binary_max = varbinary_max = 8000
    image_max = 2_147_483_647
    image_min = 1
    # endregion
    # region NotImplemented Types For Limits
    # date
    # datetime
    # smalldatetime

    # endregion


__type_based_equality = {
    "int": lambda min, max, value: min <= value <= max,
    "float": lambda min, max, value: min <= value <= max,
    "bool": lambda min, max, value: min <= value <= max,
    "str": lambda min, max, value: min <= len(value) <= max,
    "bytes": lambda min, max, value: min <= sys.getsizeof(value) <= max,
    "datetime": lambda min, max, value: True  # Not added yet

}


def _limit_check(value, data_type: str):
    min, max = getattr(__MFL__, data_type + "_min"), getattr(__MFL__, data_type + "_max")

    if __type_based_equality.get(type(value).__name__)(min, max, value):
        return value
    else:
        raise ValueError(f"value exceeds {data_type}_max value or least than {data_type}_min")


def _type_check(value, python_types: tuple, field):
    if isinstance(value, python_types):
        return value

    else:
        raise TypeError(f"Value of '{field}' field must be {', '.join([i.__name__ for i in python_types])}")


class __MFA__:

    @staticmethod
    def bigint(value, field):
        return _limit_check(_type_check(value, (int,), field), "bigint")

    @staticmethod
    def int(value, field):
        return _limit_check(_type_check(value, (int,), field), "int")

    @staticmethod
    def smallint(value, field):
        return _limit_check(_type_check(value, (int,), field), "smallint")

    @staticmethod
    def bit(value, field):
        return _limit_check(_type_check(value, (bool, int, float), field), "bit")

    @staticmethod
    def tinyint(value, field):
        return _limit_check(_type_check(value, (int,), field), "tinyint")

    @staticmethod
    def decimal(value):
        return _limit_check(_type_check(value, (float,), field), "decimal")

    @staticmethod
    def numeric(value, field):
        return _limit_check(_type_check(value, (float,), field), "numeric")

    @staticmethod
    def money(value, field):
        return _limit_check(_type_check(value, (float,), field), "money")

    @staticmethod
    def smallmoney(value, field):
        return _limit_check(_type_check(value, (float,), field), "smallmoney")

    @staticmethod
    def float(value, field):
        return _limit_check(_type_check(value, (float,), field), "float")

    @staticmethod
    def real(value, field):
        return _limit_check(_type_check(value, (float,), field), "real")

    @staticmethod
    def char(value, field):
        return _limit_check(_type_check(value, (str,), field), "char")

    @staticmethod
    def nchar(value, field):
        return _limit_check(_type_check(value, (str,), field), "nchar")

    @staticmethod
    def varchar(value, field):
        return _limit_check(_type_check(value, (str,), field), "varchar")

    @staticmethod
    def nvarchar(value, field):
        return _limit_check(_type_check(value, (str,), field), "nvarchar")

    @staticmethod
    def text(value, field):
        return _limit_check(_type_check(value, (str,), field), "text")

    @staticmethod
    def ntext(value, field):
        return _limit_check(_type_check(value, (str,), field), "ntext")

    @staticmethod
    def binary(value, field):
        return pyodbc.Binary(_limit_check(_type_check(value, (bytes,), field), "binary"))

    @staticmethod
    def nbinary(value, field):
        return pyodbc.Binary(_limit_check(_type_check(value, (bytes,), field), "nbinary"))

    @staticmethod
    def image(value, field):
        return pyodbc.Binary(_limit_check(_type_check(value, (bytes,), field), "image"))

    @staticmethod
    def date(value, field):
        return _type_check(value, (datetime.date,),field)

    @staticmethod
    def datetime(value, field):
        return _type_check(value, (datetime.datetime,),field)

    @staticmethod
    def smalldatetime(value, field):
        return _type_check(value, (datetime.time,),field)


MDC = {
    "bit": __MFA__.bit,
    "bigint": __MFA__.bigint,
    "int": __MFA__.int,
    "smallint": __MFA__.smallint,
    "tinyint": __MFA__.tinyint,
    "decimal": __MFA__.decimal,
    "numeric": __MFA__.numeric,
    "money": __MFA__.money,
    "smallmoney": __MFA__.smallmoney,
    "float": __MFA__.float,
    "real": __MFA__.real,
    "char": __MFA__.char,
    "nchar": __MFA__.nchar,
    "varchar": __MFA__.varchar,
    "nvarchar": __MFA__.nvarchar,
    "text": __MFA__.text,
    "ntext": __MFA__.ntext,
    "binary": __MFA__.binary,
    "nbinary": __MFA__.nbinary,
    "image": __MFA__.image,
    "date": __MFA__.date,
    "datetime": __MFA__.datetime,
    "smalldatetime": __MFA__.smalldatetime,

}




from msorm import models
__preview__ = False
__MAJOR__ = 1
__MINOR__ = 1
__MAINTENANCE__ = 14
__preview_version__ = f"{__MAJOR__}.{__MINOR__}.{__MAINTENANCE__}-rc2"
__version__ = f"{__MAJOR__}.{__MINOR__}.{__MAINTENANCE__}-a0"
__version__ = __preview_version__ if __preview__ else __version__
__author__ = "Mehmet Berkay Özbay"
__author_git__ = "https://github.com/bilinenkisi"
__git__ = "https://github.com/bilinenkisi/msorm"



import warnings

from msorm import settings
from msorm.exceptions import DeveloperToolsWarning

warnings.filterwarnings("once", category=DeveloperToolsWarning)

__filters__ = {
    "gt": ">",
    "gte": ">=",
    "not": "!=",
    "lt": "<",
    "lte": "<=",
    "like": " LIKE ",
    "in": " IN ",
    "not_in": " NOT IN "
}
__filtersf__ = {
    "gt": lambda value: f"'{value}'",
    "gte": lambda value: f"'{value}'",
    "not": lambda value: f"'{value}'",
    "lt": lambda value: f"'{value}'",
    "lte": lambda value: f"'{value}'",
    "like": lambda value: f"'{value}'",
    "in": lambda value: f"""({",".join(tuple(f"'{i}'" for i in value))})""",
    "not_in": lambda value: f"""({",".join(tuple(f"'{i}'" for i in value))})"""
}


class field:
    __sub_instance__ = False
    __field__ = "field"  # name of field
    __properties__ = {"default": None,
                      "null": True}

    @staticmethod
    def find_filter(field, value):
        filed_args = field.split("__")
        lenght = len(filed_args)
        if lenght > 2:
            raise ValueError(f"{field} field cannot be found in the Model")
        name = filed_args.pop(0)
        for item in filed_args:
            filter = __filters__.get(item)
            if filter:
                return name + filter + __filtersf__[item](value)
        return name + "=" + f"'{value}'"

    def __init__(self, default=None, safe=True, null=True):
        if not self.__sub_instance__ and safe:
            warnings.warn("DO NOT FORGET, USING DIRECT field CLASS IS NOT SUITABLE FOR NORMAL USAGE",
                          DeveloperToolsWarning)

        self.value = default
        self.null = null
        self.__properties__["null"] = null
        if default and safe:
            self.__properties__["default"] = self.value = settings.MDC.get(self.__field__, self.__developer_field_produce)(default)

    def __init_subclass__(cls, **kwargs):
        cls.__sub_instance__ = True

    @property
    def value(self):

        return self.__value

    @value.setter
    def value(self, val):

        self.__value = val

    @classmethod
    def get_new(cls, *args, **kwargs):
        new_field = cls(*args, **kwargs)
        return new_field

    def __developer_field_produce(self, val):
        warnings.warn("DO NOT FORGET, USING DIRECT field CLASS IS NOT SUITABLE FOR NORMAL USAGE",
                      DeveloperToolsWarning)
        return val

    def produce(self, val, field="field"):
        if val is None and self.null:
            return val

        return settings.MDC.get(self.__field__, self.__developer_field_produce)(val, field)

    def __str__(self):
        return str(self.__value)
